#!/usr/bin/env python

"""
Moin wiki format parser.

Copyright (C) 2012, 2013, 2015, 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from cgi import escape
import re

# Regular expressions.

syntax = {
    # Page regions:
    "markers"       : (r"^\s*([{]{3,}|[}]{3,})",    re.MULTILINE | re.DOTALL),  # {{{... or }}}...

    # Region contents:
    "header"        : (r"\A#!(.*?)$",               re.MULTILINE),              # #! char-excl-nl
    "region text"   : (r"(^\s*$)",                  re.MULTILINE),              # blank line
    }

# Define patterns for the regular expressions.

patterns = {}
for name, (value, flags) in syntax.items():
    patterns[name] = re.compile(value, re.UNICODE | flags)



# Document nodes.

class Container:

    "A container of document nodes."

    def __init__(self, nodes):
        self.nodes = nodes

    def append(self, node):
        self.nodes.append(node)

class Region(Container):

    "A region of the page."

    def __init__(self, nodes, level=0, type=None):
        Container.__init__(self, nodes)
        self.level = level
        self.type = type

    def have_end(self, s):
        return self.level and s.startswith("}") and self.level == len(s)

    def __repr__(self):
        return "Region(%r, %r, %r)" % (self.nodes, self.level, self.type)

    def to_string(self, out):
        out.start_region(self.level, self.type)
        for node in self.nodes:
            node.to_string(out)
        out.end_region(self.level, self.type)

class Block(Container):

    "A block in the page."

    def __init__(self, nodes, final=True):
        Container.__init__(self, nodes)
        self.final = final

    def __repr__(self):
        return "Block(%r)" % self.nodes

    def to_string(self, out):
        out.start_block(self.final)
        for node in self.nodes:
            node.to_string(out)
        out.end_block(self.final)

class Text:

    "A text node."

    def __init__(self, s):
        self.s = s

    def __repr__(self):
        return "Text(%r)" % self.s

    def to_string(self, out):
        out.text(self.s)



# Serialisation.

class Serialiser:

    "General serialisation support."

    def __init__(self, out):
        self.out = out

class MoinSerialiser(Serialiser):

    "Serialisation of the page."

    def start_region(self, level, type):
        out = self.out
        if level:
            out("{" * level)        # marker
        if type:
            out("#!%s" % type)      # header

    def end_region(self, level, type):
        out = self.out
        if level:
            out("}" * level)        # marker

    def start_block(self, final):
        pass

    def end_block(self, final):
        if not final:
            self.out("\n")

    def text(self, s):
        self.out(s)

class HTMLSerialiser(Serialiser):

    "Serialisation of the page."

    def start_region(self, level, type):
        l = []
        out = l.append
        if level:
            out("level-%d" % level)                 # marker

        # NOTE: Encode type details for CSS.

        if type:
            out("type-%s" % escape(type, True))     # header

        self.out("<span class='%s'>" % " ".join(l))

    def end_region(self, level, type):
        self.out("</span>")

    def start_block(self, final):
        self.out("<p>")

    def end_block(self, final):
        self.out("</p>")

    def text(self, s):
        self.out(escape(s))



# Parser functions.

def parse_page(s):

    """
    Parse page text 's'. Pages consist of regions delimited by markers.
    """

    region = Region([])
    items = iter(patterns["markers"].split(s))
    parse_region(items, region)
    return region

def parse_region(items, region):

    "Parse the data provided by 'items' to populate 'region'."

    nodes = region.nodes
    first = True

    # Process exposed text and sections.

    try:
        while True:

            # Parse section headers.

            if first:
                match_text = parse_region_header(items.next(), region)
                first = False
            else:
                match_text = items.next()

            # Start a section if an appropriate marker is given.

            if match_text.startswith("{"):
                _region = Region([], len(match_text))
                region.append(_region)
                parse_region(items, _region)

            # Interpret the given marker, closing the current section if the
            # given marker is the corresponding end marker for the current
            # section.

            elif region.have_end(match_text):
                return

            # Otherwise, parse text in the region.

            else:
                parse_region_text(match_text, region)

    except StopIteration:
        pass

def parse_region_header(s, region):

    """
    Parse the text 's', extracting any region header and setting it for the
    given 'region'. Return the remaining text.
    """

    items = iter(patterns["header"].split(s))
    pre_header = items.next()

    if not pre_header:
        region.type = items.next()
        return items.next()
    else:
        return pre_header

def parse_region_text(s, region):

    "Parse the text 's' as part of 'region'."

    items = iter(patterns["region text"].split(s))
    block = Block([])
    region.append(block)

    try:
        while True:
            match_text = items.next()

            if not match_text.strip():
                region.append(block)
                block.final = False
                block = Block([])
            else:
                block.append(Text(match_text))

    except StopIteration:
        pass

# Top-level parsing function.

parse = parse_page

# Top-level serialising functions.

def serialise(doc, serialiser=MoinSerialiser):
    l = []
    doc.to_string(serialiser(l.append))
    return "".join(l)

# vim: tabstop=4 expandtab shiftwidth=4
